/**
 * PilrcEdit - a graphical editor for PilRC files.
 * Copyright (C) 2000 Rob Nielsen
 * rnielsen@wn.com.au
 * http://www.wn.com.au/rnielsen/
 *
 * Released under the GNU GPL - http://www.gnu.org/copyleft/gpl.html
 */
package pilrcedit;

/**
 * This class encapsulates an arithmetic expression that evaluates to an integer value.
 * Examples might include "5", "(4+6)/2" or "X_POS+5" (where X_POS is a constant).
 */
public class Expression
{
	/** a private instance for the static evaluate method */
	private static Expression instance;

	/**
	 * Evaluates a string expression without requiring the creation of an instance.
	 * @param s the expression to evaluate
	 * @return the integer value of the expression
	 */
	public static int evaluate(String s)
	{
		if (instance==null)
		  instance=new Expression(s);
		else
		  instance.setExpression(s);
		return instance.getValue();
	}

	/** The string value of the expression, or null for a straight int value */
	String exp;

	/**
	 * The integer value of the expression - not guaranteed to be accurate
	 * unless evaluate() is called
	 */
	int value;

  /**
   * Constructs a new blank expression
   */
  public Expression()
  {
	}

  public Expression(Expression e)
  {
		setExpression(e.getExpression());
	}

	/**
	 * Constructs a new expression with the given string value
	 */
  public Expression(String s)
  {
		setExpression(s);
	}

	/**
	 * Constructs a new expression with the given integer value.
	 */
	public Expression(int i)
	{
		setValue(i);
	}

	/**
	 * Is this expression a straight integer value?
	 */
	public boolean isAbsolute()
	{
		return exp==null;
	}

  /**
   * Is this expression set to the empty string?
   */
  public boolean isEmpty()
  {
		return exp!=null&&exp.length()==0;
	}

  /**
   * Returns a string representation of this object.  Same result as calling
   * getExpression()
   */
  public String toString()
  {
		return getExpression();
  }

  /**
   * Returns the string representation of this expression.
   */
  public String getExpression()
  {
		if (exp==null)
		  return ""+value;
		return exp;
	}

  /**
   * Sets the expression string for this expression.
   * @param s the string to set
   */
  public void setExpression(String s)
  {
		exp=s;
		evaluate();
	}

  /**
   * Gets the integer value of this expression.
   */
  public int getValue()
  {
		return value;
	}

  /**
   * Sets this expression as a straight integer value.  Will remove
   * any existing expression.
   */
  public void setValue(int i)
  {
		exp=null;
		value=i;
		evaluate();
	}

	/**
	 * Evaluates the expression string and returns the result (which is also available
	 * with getValue()).  Any parse errors will throw a IllegalArgumentException
	 * @return the integer value of the expression
	 * @throws java.lang.IllegalArgumentException
	 */
	public int evaluate()
	{
		if (exp==null)
		  return value;
		if (exp.length()==0)
		{
		  return (value=0);
		}
		try
		{
			value=Integer.parseInt(exp);
			exp=null;
			return value;
		}
		catch(NumberFormatException e)
		{
			return (value=evaluate(exp.toCharArray(),0,exp.length()));
		}
	}

	/**
	 * Recursively evaluates the exp contained in the char array e between
	 * start and end.
	 * @param e the exp
	 * @param start the start index (inclusive) to evaluate from
	 * @param end the end index (exclusive) to evaluate to
	 * @return the evaluated exp
	 * @throws java.lang.IllegalArgumentException if the exp can't be parsed.
	 */
	private int evaluate(char[] e,int start,int end)
	{
		if (end==start)
			return 0;
		//System.out.println("evaluate("+start+","+end+")");
		boolean number=true;
		int bracket=0;
		if (e[end-1]==')'&&e[start]=='(')
		{
			start++;
			end--;
			if (end==start)
				return 0;
		}
		for(int i=end-1;i>=start;i--)
		{
			if (e[i]==')')
				bracket++;
			else
			if (e[i]=='(')
				bracket--;
			else
			if (bracket==0)
			{
				if (e[i]=='+')
					return evaluate(e,start,i)+evaluate(e,i+1,end);
				if (e[i]=='-')
					return evaluate(e,start,i)-evaluate(e,i+1,end);
				if (e[i]=='*')
					return evaluate(e,start,i)*evaluate(e,i+1,end);
				if (e[i]=='/')
					return evaluate(e,start,i)/evaluate(e,i+1,end);
			}
			if (e[i]<'0'||e[i]>'9')
				number=false;
		}
		String s=new String(e,start,end-start);
		if (number)
			return Integer.parseInt(s);
		return evaluateString(s);
	}

  /**
   * Evaluates a string found in the expression which isn't an arithmetic
   * expression.  Sublasses should override this method to add functionality
   * to expressions but should contain <code>return super.evaluateString(s);</code>
   * as their last line.
   */
  protected int evaluateString(String s)
  {
	  Project project=Project.getProject();
	  if (project.hasConstant(s))
			return project.getConstant(s);
		throw new IllegalArgumentException("Error parsing expression: \""+s+"\"");
	}

  /**
   * Modifies the given string exp by the given amount.
   * For example modifyExpression("PREVLEFT+2,3) would return
   * "PREVLEFT+5".  Used when dragging controls around.
   * @param s the string to modify (usually loc[x])
   * @param amnt the amount to change by
   * @return the modified exp, or null if the absolute value should be used
   */
  public int modifyExpression(int amnt)
	{
		if (amnt==0)
		  return 0;
		value+=amnt;
		if (exp==null)
		  return 0;

		boolean possiblePlus=false;
		boolean possibleMinus=false;

		// look for blah+x
		int ind=exp.lastIndexOf('+');
		if (ind!=-1)
		{
			possiblePlus=true;
			try
			{
				int i=Integer.parseInt(exp.substring(ind+1));
				i+=amnt;
				if (i<0)
					exp=exp.substring(0,ind)+i;
				else
				if (i==0)
					exp=exp.substring(0,ind);
				else
					exp=exp.substring(0,ind)+"+"+i;
			  return i;
			}
			catch(NumberFormatException e)
			{
			}
		}

		// look for blah-x
		ind=exp.lastIndexOf('-');
		if (ind!=-1)
		{
			possibleMinus=true;
			try
			{
				int i=Integer.parseInt(exp.substring(ind));
				i+=amnt;
				if (i<0)
					exp=exp.substring(0,ind)+i;
				else
				if (i==0)
					exp=exp.substring(0,ind);
				else
					exp=exp.substring(0,ind)+"+"+i;
				return i;
			}
			catch(NumberFormatException e)
			{
			}
		}

		// look for x+blah
		if (possiblePlus)
		{
  		ind=exp.indexOf('+');
			try
			{
				int i=Integer.parseInt(exp.substring(0,ind));
				i+=amnt;
				if (i==0)
					exp=exp.substring(ind+1);
				else
					exp=i+exp.substring(ind);
			  return 0;
			}
			catch(NumberFormatException e)
			{
			}
	  }

	  // look for x-blah
		if (possibleMinus)
		{
  		ind=exp.indexOf('-');
			try
			{
				int i=Integer.parseInt(exp.substring(0,ind));
				i+=amnt;
				if (i==0)
					exp=exp.substring(ind+1);
				else
					exp=i+exp.substring(ind);
			  return 0;
			}
			catch(NumberFormatException e)
			{
			}
		}

		// ok, we haven't found a match so it's just something straight like PREVLEFT
		// so tack it on the end
		if (amnt>0)
			exp+="+"+amnt;
		else
		if (amnt<0)
			exp+=amnt;
	  return amnt;
	}
}
