/**
 * PilrcEdit - a graphical editor for PilRC files.
 * Copyright (C) 2000 Rob Nielsen
 * rnielsen@wn.com.au
 * http://www.wn.com.au/rnielsen/
 *
 * Released under the GNU GPL - http://www.gnu.org/copyleft/gpl.html
 */
package pilrcedit;

import java.io.*;
import java.util.*;
import java.awt.Image;

public class Project extends Resource
{
  // changelistener constants
  public static final int DIRTY=0;
  public static final int CLEAN=1;
  public static final int FORM_ADDED=2;
  public static final int FORM_ADDED_JUMPTO=3;
  public static final int FORM_MODIFIED=4;
  public static final int FORM_DELETED=5;
  public static final int CONSTANTS_CHANGED=6;
  public static final int LOADED_FILE=7;

	static Project project;

  ResourceCollection includes=new ResourceCollection(null);
  ResourceCollection fonts=new ResourceCollection("Fonts");
  ResourceCollection forms=new ResourceCollection("Forms and Alerts",true);
  ResourceCollection bitmaps=new ResourceCollection("Bitmaps and Icons");
  ResourceCollection strings=new ResourceCollection("Strings");
  Vector unsupported=new Vector();

  String file;
  Hashtable idMap;
  String parentDir;
  String comments;
  boolean dirty=false;
  boolean parsing=false;
  int currentConstantFile=0;
  String ifdef;
  String ifndef;
  int commentLines;

  public static Project getProject()
  {
		return project;
	}

  public Project()
  {
		project=this;
	}

  public void newProject()
  {
		file=null;
		reset();
		newForm();
	}

  public void setDirty(boolean b)
  {
		//System.out.println("setDirty("+b+") - "+dirty);
		if (dirty!=b)
		{
			dirty=b;
      fireObjectChanged(dirty?DIRTY:CLEAN);
		}
	}

	public boolean isDirty()
	{
		return dirty;
	}

  public ConstantFile getConstantFile(int i)
  {
		return (ConstantFile)includes.getResourceAt(i);
	}

  public Enumeration getConstantFiles()
  {
		return includes.elements();
	}

  public boolean hasConstant(String id)
  {
		if (includes.size()==0)
		{
			if (isParsing())
			{
				ConstantFile cf=new ConstantFile();
				cf.setPilRCGenerated(true);
				includes.add(cf);
				return true;
			}
		  return false;
		}
		for(Enumeration e=includes.elements();e.hasMoreElements();)
		  if (((ConstantFile)e.nextElement()).hasConstant(id))
		    return true;
		return false;
	}

  public int getConstant(String id)
  {
		if (includes==null)
		  return -1;
		for(Enumeration e=includes.elements();e.hasMoreElements();)
		{
			ConstantFile cf=(ConstantFile)e.nextElement();
		  if (cf.hasConstant(id))
      {
        int i=cf.getConstant(id);
        if (cf.isAutogenerating())
          fireObjectChanged(CONSTANTS_CHANGED);
		    return i;
      }
		}
		return -1;
	}

  public int getConstantCount()
  {
		return getCurrentConstantFile().getConstantCount();
	}

  public Enumeration getConstantKeys()
  {
		return getCurrentConstantFile().getAllConstants();
	}

  public void removeConstant(String s)
  {
		getCurrentConstantFile().removeConstant(s);
	}

  public void setConstant(String s,int value)
  {
		setConstant(s,value,true);
	}

  public void setConstant(String s,int value,boolean updateFrame)
  {
		getCurrentConstantFile().setConstant(s,value);
		if (updateFrame)
      fireObjectChanged(CONSTANTS_CHANGED);
	}

  public void addID(String id)
  {
		getCurrentConstantFile().addID(id);
    fireObjectChanged(CONSTANTS_CHANGED);
	}

  public ConstantFile setCurrentConstantFile(int i)
  {
		currentConstantFile=i;
		return getCurrentConstantFile();
	}

	public ConstantFile getCurrentConstantFile()
	{
		if (includes.size()==0)
		  includes.add(new ConstantFile());
		return (ConstantFile)includes.getResourceAt(currentConstantFile);
	}

  public void open(String file)
  {
		this.file=file;
		parse(file);
	}

	public Form reload(Form f)
	{
		if (file!=null)
		{
			int pos=forms.getIndexOf(f);
			parse(file);
			if (pos!=-1)
  			return (Form)forms.getResourceAt(pos);
  	  else
  	    return null;
		}
		return f;
	}

  public Form getFirstForm()
  {
		for(int i=0,size=forms.size();i<size;i++)
		{
			Form f=(Form)forms.getResourceAt(i);
		  if (f!=null)
		    return f;
		}
		return null;
	}

  public int getNumForms()
  {
    return forms.size();
  }

  public Form newForm()
  {
		Form form=new Form(this);
    addForm(form);
		return form;
	}

	public Alert newAlert()
	{
		Alert alert=new Alert(this);
		addForm(alert);
		return alert;
	}

  public void addForm(Form form)
  {
    addForm(form,false);
  }

  public void addForm(Form form,boolean jumpTo)
  {
    forms.add(form);
    fireObjectChanged(jumpTo?FORM_ADDED_JUMPTO:FORM_ADDED,form);
  }

  public void deleteForm(Form f)
  {
    forms.remove(f);
    fireObjectChanged(FORM_DELETED,f);
	}

  public Form getFormAtIndex(int id)
  {
		return (Form)forms.getResourceAt(id);
	}

	public String getString(int id)
	{
		return ((StringResource)strings.getResource(id)).getString();
	}

	public Bitmap getBitmap(int id)
	{
    return (Bitmap)bitmaps.getResource(id);
    /*
    if (bitmap==null)
    {
      //System.out.println("getBitmap("+id+")=null");
      return null;
    }

		return bitmap;*/
	}

  public String getParentDir()
  {
		return parentDir;
	}

  /**
   * Creates a File object from a string which is either an absolute path, or
   * a relative path from the location of this project's .rcp
   * @param file the filename
   * @return the corresponding File object
   */
  public File makeRelativeFile(String file)
  {
    File f=new File(file);
    if (!f.isAbsolute()&&parentDir!=null)
      f=new File(parentDir,file);
    return f;
  }

  /**
   * For files like bitmaps, strings and includes, converts an absolute
   * string (eg c:\projects\blah\images\pic.bmp) to one relative to the
   * dir this projects .rcp file is in (eg images\pic.bmp).  If the file
   * is already relative or not in a relative position, the string is
   * returned unchanged.
   * @param file the file to make relative
   * @return the relative version
   */
  public String convertToRelativeFilename(String file)
  {
    if (parentDir==null)
      return file;
    String par=parentDir.toLowerCase();
    String fl=file.toLowerCase();
    if (fl.startsWith(par))
    {
      file=file.substring(par.length());
    }
    return file;
  }

  public void titleChanged(Form f)
  {
    fireObjectChanged(FORM_MODIFIED,f);
	}

  public ResourceCollection getBitmaps()
  {
    while (bitmaps.size()<2)
      bitmaps.add((Resource)null);
    return bitmaps;
  }

  public void reset()
  {
		setDirty(false);
    fonts.clear();
    forms.clear();
    includes.clear();
    strings.clear();
    bitmaps.clear();
    unsupported.removeAllElements();
		ConstantFile.resetAutogen(1000);
    fireObjectChanged(ChangeListener.CLEARED);
    parentDir=null;
    comments=null;
	}

  public boolean isParsing()
  {
		return parsing;
	}

  public String getIfDef()
  {
    return ifdef;
  }

  public String getIfNDef()
  {
    return ifndef;
  }

  public void parse(String file)
  {
		parsing=true;
		reset();
		StreamTokenizer st=null;
		try
		{
			File f=new File(new File(file).getAbsolutePath());
			parentDir=f.getParent()!=null?f.getParent()+File.separator:"";
			BufferedReader br=new BufferedReader(new FileReader(f));
			st=new StreamTokenizer(br);
			st.resetSyntax();
			st.whitespaceChars(0,32);
			st.wordChars(33,255);
			st.ordinaryChar('/');
			st.slashSlashComments(true);
      st.slashStarComments(true);
			st.lowerCaseMode(true);

      parse(st);

      br.close();
      br=new BufferedReader(new FileReader(f));
      StringBuffer sb=new StringBuffer();
      for(int i=1;i<commentLines;i++)
      {
        sb.append(br.readLine()).append(Preferences.newLine);
      }
      comments=sb.toString();
      br.close();
      /*
      if (autoGenerate)
      {
        autoGenerate=false;
        writeIncludeFile();
      }
      */
      if (includes!=null)
      {
        for(Enumeration e=includes.elements();e.hasMoreElements();)
        {
          ConstantFile cf=(ConstantFile)e.nextElement();
          cf.setAutoGenerate(false);
        }
      }

      parsing=false;

      fireObjectChanged(LOADED_FILE,file);
    }
    catch(Exception e)
    {
      GeneralDialog gd=Preferences.getDialog();
      gd.setTitle("Parse Error");
      gd.setText("PilrcEdit has encountered an error parsing \""+file+"\""+
        (st!=null?" at line "+st.lineno():"")+". If PilRC accepts this file, then "+
        "it indicates a bug in PilrcEdit.  Please send this file (along with any"+
        "associated files like .h's) to rnielsen@wn.com.au and I'll fix it.");
      gd.setButtonOk();
      gd.show();
      if (st!=null)
        System.out.println("st.lineno="+st.lineno()+" st.sval=\""+st.sval+"\" st.ttype="+st.ttype);
      //e.printStackTrace();
      Debug.debug(e);
    }
  }

  public void parse(StreamTokenizer st) throws IOException
  {
    commentLines=-1;
    int tok=0;
    while ((tok=st.nextToken())!=StreamTokenizer.TT_EOF)
    {
      if (tok==StreamTokenizer.TT_WORD)
      {
        if (commentLines==-1)
        {
          commentLines=st.lineno();
          st.wordChars('/','/');
        }

        if (st.sval.equals("#include"))
        {
          String s=readQuotedString(st);
          includes.add(new ConstantFile(parentDir,s));
        }
        else
        if (st.sval.equals("#ifdef"))
        {
          ifdef=readNormalString(st);
        }
        else
        if (st.sval.equals("#ifndef"))
        {
          ifndef=readNormalString(st);
        }
        else
        if (st.sval.equals("#endif"))
        {
          ifdef=null;
          ifndef=null;
        }
        else
        if (st.sval.equals("form"))
        {
          Form form=new Form(this,st);
          addForm(form);
        }
        else
        if (st.sval.equals("alert"))
        {
          Alert alert=new Alert(this,st);
          addForm(alert);
        }
        else
        if (st.sval.startsWith("bitmap")||
            st.sval.startsWith("icon")||
            st.sval.startsWith("smallicon"))
        {
          Bitmap bitmap=new Bitmap(this,st);
          switch(bitmap.getType())
          {
            case Bitmap.ICON:
              bitmaps.set(0,bitmap);
              break;
            case Bitmap.SMALLICON:
              bitmaps.set(1,bitmap);
              break;
            case Bitmap.BITMAP:
              while (bitmaps.size()<2)
                bitmaps.add((Resource)null);
              bitmaps.add(bitmap);
              break;
          }
        }
        else
        if (st.sval.equals("font"))
        {
          String fid=parseID(st);
          st.nextToken();
          st.nextToken();
          Expression font=new Expression(st.sval);
          int fontID=font.getValue();
          String fontfile=readQuotedString(st);
          Font.registerFont(fontID,parentDir+fontfile);
          fonts.add("FONT ID "+fid+" FONTID "+font+" \""+fontfile+"\"");
        }
        else
        if (st.sval.equals("menu"))
        {
          StringBuffer sb=new StringBuffer("MENU");
          st.ordinaryChar(' ');
          st.ordinaryChar('\t');
          st.wordChars(' ',' ');
          st.wordChars('\t','\t');
          st.lowerCaseMode(false);
          st.nextToken();
          sb.append(st.sval);
          int cnt=0;
          while(true)
          {
            if (st.sval.trim().equalsIgnoreCase("BEGIN"))
              cnt++;
            else
            if (st.sval.trim().equalsIgnoreCase("END"))
            {
              cnt--;
              if (cnt==0)
                break;
            }
            st.nextToken();
            sb.append(Preferences.newLine).append(st.sval);
          }
          st.lowerCaseMode(true);
          st.whitespaceChars(' ',' ');
          st.whitespaceChars('\t','\t');
          unsupported.addElement(sb.toString());
        }
        else
        if (st.sval.equals("string"))
        {
          strings.add(new StringResource(this,st));
        }
        else
        {
          // unsupported
          boolean isComment=st.sval.startsWith("//");
          st.eolIsSignificant(true);
          st.lowerCaseMode(false);
          StringBuffer sb=new StringBuffer(st.sval.toUpperCase());
          while(st.nextToken()==st.TT_WORD)
            sb.append(' ').append(st.sval);
          if (!isComment)
            unsupported.addElement(sb.toString());
          st.lowerCaseMode(true);
          st.eolIsSignificant(false);
        }
      }
    }
  }

  public void setParentDir(String pd)
  {
    if (pd.equals(parentDir))
      return;
    parentDir=pd;
    for(Enumeration e=bitmaps.elements();e.hasMoreElements();)
    {
      Bitmap b=(Bitmap)e.nextElement();
      if (b!=null)
        b.parentDirChanged();
    }
  }

  public void save()
  {
		save(file);
	}

  public void save(String f)
  {
		if (f!=null)
		  file=f;
		else
		if (file!=null)
			(new File(file)).renameTo(new File(file.substring(0,file.length()-4)+"bak.rcp"));

		try
		{
      File fi=new File(new File(file).getAbsolutePath());
      setParentDir(fi.getParent()!=null?fi.getParent()+File.separator:"");

			PrintStream ps=new PrintStream(new FileOutputStream(file));
      write(ps);
			ps.close();
			setDirty(false);
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
	}

  public void write(PrintStream ps)
  {
    Keywords.setCase(Keywords.PREFERENCES);
    if (comments!=null)
      ps.print(comments);

    for(Enumeration e=includes.elements();e.hasMoreElements();)
    {
      ConstantFile cf=(ConstantFile)e.nextElement();
      if (cf.isDirty())
      {
        if (cf.getFile()==null)
        {
          File fi=new File(file);
          String parent=fi.getParent();
          String s=fi.getName();
          int i=s.lastIndexOf('.');
          cf.setFilename(parent,s.substring(0,i)+".h");
        }
        cf.write();
      }
      ps.println("#include \""+cf.getFile()+"\"");
    }
    ps.println();

    fonts.write(ps);
    forms.write(ps);
    bitmaps.write(ps);
    strings.write(ps);
    for(Enumeration e=unsupported.elements();e.hasMoreElements();)
    {
      ps.println(e.nextElement());
      ps.println();
    }
    Keywords.setCase(Keywords.MIXED_CASE);
  }
}